# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""Eelbrain-specific wx IDs"""
import wx

APYDOC_NEW = wx.NewIdRef()
ATTACH = wx.NewIdRef()

BASELINE_CUSTOM = wx.NewIdRef()
BASELINE_GLOABL_MEAN = wx.NewIdRef()
BASELINE_NONE = wx.NewIdRef()

CANVAS_PANEL_ZOOM = wx.NewIdRef()
CANVAS_PANEL_OVERVIEW = wx.NewIdRef()
CLEAR_CACHE = wx.NewIdRef()
CLEAR_TERMINAL = wx.NewIdRef()
COLOUR_CHOOSER = wx.NewIdRef()
COMMENT = wx.NewIdRef()
COPY_AS_PNG = wx.NewIdRef()

DATASET_ATTACH = wx.NewIdRef()
DATASET_IMPORT = wx.NewIdRef()
DRAW_CROSSHAIRS = wx.NewIdRef()
DUPLICATE = wx.NewIdRef()
DUPLICATE_WITH_OUTPUT = wx.NewIdRef()

EXPERIMENT_LOAD = wx.NewIdRef()
EXPERIMENT_NEW = wx.NewIdRef()
ENABLE_STARTUP_SCRIPT = wx.NewIdRef()
EXEC_DOCUMENT = wx.NewIdRef()
EXEC_DOCUMENT_FROM_DISK = wx.NewIdRef()
EXEC_FILE = wx.NewIdRef()
EXEC_SELECTION = wx.NewIdRef()

FIG_TITLE = wx.NewIdRef()
FOCUS_SHELL = wx.NewIdRef()

GRAND_AVERAGE = wx.NewIdRef()

HELP_EELBRAIN = wx.NewIdRef()
HELP_HIDE = wx.NewIdRef()
HELP_MDP = wx.NewIdRef()
HELP_MPL = wx.NewIdRef()
HELP_PDB = wx.NewIdRef()
HELP_PYTHON = wx.NewIdRef()

INSERT_Color = wx.NewIdRef()
INSERT_Path_file = wx.NewIdRef()
INSERT_Path_dir = wx.NewIdRef()
INSERT_Path_new = wx.NewIdRef()

LINK_TIME_AXES = wx.NewIdRef()

OPEN = wx.NewIdRef()
OPEN_HISTORY = wx.NewIdRef()
OPEN_HISTORY_CURRENT = wx.NewIdRef()
OPEN_HISTORY_PAST = wx.NewIdRef()

PLOT_COLORBAR = wx.NewIdRef()
PYDOC_NEW = wx.NewIdRef()
PYDOC_LOAD = wx.NewIdRef()
PYDOC_EXEC_ISOLATE = wx.NewIdRef()
PYDOC_HASH = wx.NewIdRef()
PYPLOT_CLOSEALL = wx.NewIdRef()
PYPLOT_DRAW = wx.NewIdRef()

P_MGR = wx.NewIdRef()
PCA_REMOVE = wx.NewIdRef()

RECENT_LOAD_E = wx.NewIdRef()
RECENT_LOAD_PYDOC = wx.NewIdRef()
REDO = wx.NewIdRef()
RESIZE = wx.NewIdRef()

SAVE_REJECTION = wx.NewIdRef()
SET_BAD_CHANNELS = wx.NewIdRef()
SET_FONT = wx.NewIdRef()
SET_LAYOUT = wx.NewIdRef()
SET_MARKED_CHANNELS = wx.NewIdRef()
SET_TEND = wx.NewIdRef()
SET_TITLE = wx.NewIdRef()
SET_TSTART = wx.NewIdRef()
SET_VLIM = wx.NewIdRef()
SHELL_Clear = wx.NewIdRef()
SHOW_EXAMPLES = wx.NewIdRef()
SHOW_SOURCES = wx.NewIdRef()
SIZE_MAX = wx.NewIdRef()
SIZE_MAX_NOTITLE = wx.NewIdRef()
SIZE_MIN = wx.NewIdRef()
SMOOTHING = wx.NewIdRef()
SPLITTER = wx.NewIdRef()
SPLITTER2 = wx.NewIdRef()

TABLE = wx.NewIdRef()
TABLE_COL = wx.NewIdRef()
TABLE_ROW = wx.NewIdRef()
TABLE_SAVE = wx.NewIdRef()
THRESHOLD = wx.NewIdRef()
TIME = wx.NewIdRef()
TOOLS = wx.NewIdRef()

UNDO = wx.NewIdRef()

VIEW_LATERAL = wx.NewIdRef()
VIEW_MEDIAL = wx.NewIdRef()

WINDOW_MINIMIZE = wx.NewIdRef()
WINDOW_TILE = wx.NewIdRef()
WINDOW_ZOOM = wx.NewIdRef()

YIELD_TO_TERMINAL = wx.NewIdRef()
